/* 
 * $Id: config.c,v 1.11 1998/07/23 17:57:20 labovit Exp $
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>

#include <mrt.h>
#include <trace.h>
#include <io.h>
#include <select.h>
#include <stdarg.h>
#include <interface.h>
#include <bgp.h>
#include <rib.h>
#include <config_file.h>
#include "config.h"
#include "bgpsim.h"

static config_bgpsim_t *CONFIG_BGPSIM;

int config_router_neighbor_stability (uii_connection_t *uii,
				      prefix_t *prefix, int stability);
static network_t *
config_find_network (int num)
{
    network_t *network;

    LL_Iterate (SIMULATION->ll_networks, network) {
	if (network->num == num)
	    return (network);
    }

    return (NULL);
}


static void
get_config_network_list (int num)
{
    network_t *network;
    range_t *range;
    int i;

	if ((network = config_find_network (num)) == NULL)
		return;

    config_add_output ("network-list %d\n", num);
    LL_Iterate (network->ll_range, range) {
	config_add_output ("  range %s %s\n", prefix_toax (range->start), 
					      prefix_toa (range->end));
    }
    config_add_output ("  stability %d\n", network->stability);
    config_add_output ("  change %d\n", network->change_interval);
    config_add_output ("  route-map");
    for (i = 1; i <= network->max_set; i++) {
	config_add_output (" %d", network->sets[i]);
    }
    config_add_output ("\n");
}


static int 
config_network_list (uii_connection_t * uii, int num)
{
    network_t *network;

	if ((network = config_find_network (num)) == NULL) {
            network = New (network_t);
	    network->num = num;
	    memset (network->sets, 0, sizeof (network->sets));
	    LL_Add (SIMULATION->ll_networks, network);
	    CONFIG_BGPSIM->network = network;
	    uii->state = UII_CONFIG_NETWORK_LIST;
	    config_add_module (0, "network-list", get_config_network_list,
			      (void *)num);
	    return (1);
        }
	else {
	    trace (ERROR, CONFIG.trace,
		   "*ERROR* CONFIG network-list %d already defined\n", num);
    	    return (-1);
	}
}


static int
network_list_range (uii_connection_t * uii, prefix_t *start, prefix_t *end)
{
	range_t *range;

	if (start->family != end->family) {
	    Deref_Prefix (start);
	    Deref_Prefix (end);
	    return (-1);
	}

	if (CONFIG_BGPSIM->network->ll_range == NULL)
            CONFIG_BGPSIM->network->ll_range = LL_Create (0);

	range = New (range_t);
	range->start = start;
	range->end = end;
	LL_Add (CONFIG_BGPSIM->network->ll_range, range);
	return (1);
}


static int
network_list_stability (uii_connection_t * uii, int stability)
{
    char tmpx[MAXLINE];

    /* stability */
        sprintf (tmpx, "Network Stability %d", CONFIG_BGPSIM->network->num);
	CONFIG_BGPSIM->network->stability = stability;
	CONFIG_BGPSIM->network->timer_stability =
	New_Timer (network_schedule_stability,
		   CONFIG_BGPSIM->network->stability, tmpx, 
		   CONFIG_BGPSIM->network);
	return (1);
}

static int network_list_load_file (uii_connection_t *uii, char *filename) {
  
  /* don't allow both range and filename to be configured */

  /*if (CONFIG_BGPSIM->network->ll_routes == NULL)
    CONFIG_BGPSIM->network->ll_routes = LL_Create (0);*/

  CONFIG_BGPSIM->network->filename = filename;

  return (1);
}


static int
network_list_change (uii_connection_t * uii, int change)
{
    char tmpx[MAXLINE];

	sprintf (tmpx, "Network Change %d", CONFIG_BGPSIM->network->num);
	CONFIG_BGPSIM->network->change_interval = change;
	CONFIG_BGPSIM->network->timer_change = 
		New_Timer (network_schedule_change,
			   CONFIG_BGPSIM->network->change_interval, tmpx,
			   CONFIG_BGPSIM->network);
	return (1);
}


static int
network_list_set (uii_connection_t * uii, char *arg)
{
    char *list = arg;
    char tmpx[MAXLINE], *cp;

	int set;

	while ((cp = uii_parse_line2 (&list, tmpx)) != NULL) {

	    if (!isdigit (cp[0])) {
		Delete (arg);
		return (0);
	    }

	    set = atoi (cp);

	    if (set <= 0) {
		trace (ERROR, CONFIG.trace, "CONFIG set number should be > 0\n");
		Delete (arg);
		return (-1);
	    }
	    if (CONFIG_BGPSIM->network->max_set >= MAX_SETS - 1) {
		trace (ERROR, CONFIG.trace, 
		       "CONFIG too many sets %d (should be < %d)\n", 
		       CONFIG_BGPSIM->network->max_set, MAX_SETS);
		Delete (arg);
		return (-1);
	    }
	    CONFIG_BGPSIM->network->sets[++CONFIG_BGPSIM->network->max_set] 
		= set;
	    trace (NORM, CONFIG.trace, 
		   "CONFIG new set %d for network-list %d\n", 
		   set, CONFIG_BGPSIM->network->num);
	}
	Delete (arg);
	return (1);
}


#ifdef notdef
static int
route_set_compare (u_int *a, u_int *b)
{
   return (*a - *b);
}
#endif








void 
add_bgpsim_config (trace_t * trace) {

    CONFIG_BGPSIM = New (config_bgpsim_t);

    uii_add_command2 (UII_CONFIG, COMMAND_NORM, "network-list %d", 
		config_network_list, "Defines a network list");

    uii_add_command2 (UII_CONFIG_NETWORK_LIST, COMMAND_NORM, "range %m %m", 
		     network_list_range, "Prefix range to generate");
    uii_add_command2 (UII_CONFIG_NETWORK_LIST, COMMAND_NORM, "stability %d", 
		     network_list_stability, "Flapping interval (sec)");
    uii_add_command2 (UII_CONFIG_NETWORK_LIST, COMMAND_NORM, "change %d", 
		     network_list_change, "Rotates route-maps");
    uii_add_command2 (UII_CONFIG_NETWORK_LIST, COMMAND_NORM, "route-map %S",
		      network_list_set, "Lists route-map to use");
    uii_add_command2 (UII_CONFIG_NETWORK_LIST, COMMAND_NORM, "file %sfilename",
		      network_list_load_file, 
		      "Load routes from routing table dump");
    uii_add_command2 (UII_CONFIG_NETWORK_LIST, COMMAND_NORM, "file %sfilename",
		      network_list_load_file, 
		      "Load routes from routing table dump");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %m stability %d", 
		      config_router_neighbor_stability,
		      "Set stability of TCP peering session with neighbor");
}

#include "../mrtd/config.c"




/* neighbor xx.xx.xx.xx stability %d */
int config_router_neighbor_stability (uii_connection_t *uii,
				      prefix_t *prefix, int stability) {
  bgp_peer_t *peer;
  peer_flap_t *flap;

  if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
    config_notice (TR_ERROR, uii, "No peer %s\n", prefix_toa (prefix));
    Deref_Prefix (prefix);
    return (-1);
  }
  Deref_Prefix (prefix);

  if (SIMULATION->ll_peer_flaps == NULL)
    SIMULATION->ll_peer_flaps =  LL_Create (0);

  flap = New (peer_flap_t);
  flap->peer = peer;
  flap->timer = New_Timer (bgpsim_flap_peer, stability, "Flap Peer", peer);

  LL_Add (SIMULATION->ll_peer_flaps, flap);

  return (1);
}
